@@ -10,13 +10,12 @@ from concurrent.futures import Future
 from awscrt import NativeResource
 from awscrt.http import HttpRequest
 from awscrt.io import ClientBootstrap, TlsConnectionOptions
-from awscrt.auth import AwsCredentials, AwsCredentialsProvider, AwsSignatureType, AwsSignedBodyHeaderType, AwsSignedBodyValue, AwsSigningAlgorithm, AwsSigningConfig
 from awscrt.auth import AwsCredentialsProvider, AwsSignatureType, AwsSignedBodyHeaderType, AwsSignedBodyValue, \
     AwsSigningAlgorithm, AwsSigningConfig
 import awscrt.exceptions
 import threading
 from dataclasses import dataclass
-from typing import List, Optional, Tuple
+from typing import List, Optional, Tuple, Sequence
 from enum import IntEnum
 
 
@@ -182,13 +181,14 @@ class S3Client(NativeResource):
             for each connection, unless `tls_mode` is :attr:`S3RequestTlsMode.DISABLED`
 
         part_size (Optional[int]): Size, in bytes, of parts that files will be downloaded or uploaded in.
-            Note: for :attr:`S3RequestType.PUT_OBJECT` request, S3 requires the part size greater than 5 MiB.
-            (8*1024*1024 by default)
+            Note: for :attr:`S3RequestType.PUT_OBJECT` request, client will adjust the part size to meet the service limits.
+            (max number of parts per upload is 10,000, minimum upload part size is 5 MiB)
 
         multipart_upload_threshold (Optional[int]): The size threshold in bytes, for when to use multipart uploads.
+            This only affects :attr:`S3RequestType.PUT_OBJECT` request.
             Uploads over this size will use the multipart upload strategy.
             Uploads this size or less will use a single request.
-            If not set, `part_size` is used as the threshold.
+            If not set, maximal of `part_size` and 5 MiB will be used.
 
         throughput_target_gbps (Optional[float]): Throughput target in
             Gigabits per second (Gbps) that we are trying to reach.
@@ -204,6 +204,15 @@ class S3Client(NativeResource):
             client can use for buffering data for requests.
             Default values scale with target throughput and are currently
             between 2GiB and 8GiB (may change in future)
+
+        network_interface_names: (Optional[Sequence(str)]) 
+            **THIS IS AN EXPERIMENTAL AND UNSTABLE API.**
+            A sequence of network interface names. The client will distribute the
+            connections across network interfaces. If any interface name is invalid, goes down,
+            or has any issues like network access, you will see connection failures.
+            This option is only supported on Linux, MacOS, and platforms that have either SO_BINDTODEVICE or IP_BOUND_IF. It
+            is not supported on Windows. `AWS_ERROR_PLATFORM_NOT_SUPPORTED` will be raised on unsupported platforms. On
+            Linux, SO_BINDTODEVICE is used and requires kernel version >= 5.7 or root privileges.
     """
 
     __slots__ = ('shutdown_event', '_region')
@@ -221,7 +230,8 @@ class S3Client(NativeResource):
             multipart_upload_threshold=None,
             throughput_target_gbps=None,
             enable_s3express=False,
-            memory_limit=None):
+            memory_limit=None,
+            network_interface_names: Optional[Sequence[str]] = None):
         assert isinstance(bootstrap, ClientBootstrap) or bootstrap is None
         assert isinstance(region, str)
         assert isinstance(signing_config, AwsSigningConfig) or signing_config is None
@@ -234,6 +244,7 @@ class S3Client(NativeResource):
             throughput_target_gbps,
             float) or throughput_target_gbps is None
         assert isinstance(enable_s3express, bool) or enable_s3express is None
+        assert isinstance(network_interface_names, Sequence) or network_interface_names is None
 
         if credential_provider and signing_config:
             raise ValueError("'credential_provider' has been deprecated in favor of 'signing_config'.  "
@@ -269,6 +280,10 @@ class S3Client(NativeResource):
             throughput_target_gbps = 0
         if memory_limit is None:
             memory_limit = 0
+        if network_interface_names is not None:
+            # ensure this is a list, so it's simpler to process in C
+            if not isinstance(network_interface_names, list):
+                network_interface_names = list(network_interface_names)
 
         self._binding = _awscrt.s3_client_new(
             bootstrap,
@@ -283,6 +298,7 @@ class S3Client(NativeResource):
             throughput_target_gbps,
             enable_s3express,
             memory_limit,
+            network_interface_names,
             s3_client_core)
 
     def make_request(
@@ -296,6 +312,8 @@ class S3Client(NativeResource):
             signing_config=None,
             credential_provider=None,
             checksum_config=None,
+            part_size=None,
+            multipart_upload_threshold=None,
             on_headers=None,
             on_body=None,
             on_done=None,
@@ -311,10 +329,25 @@ class S3Client(NativeResource):
             request (HttpRequest): The overall outgoing API request for S3 operation.
                 If the request body is a file, set send_filepath for better performance.
 
-            operation_name(Optional[str]): Optional S3 operation name (e.g. "CreateBucket").
-                This will only be used when `type` is :attr:`~S3RequestType.DEFAULT`;
-                it is automatically populated for other types.
+            operation_name(Optional[str]): S3 operation name (e.g. "CreateBucket").
+                This MUST be set when `type` is :attr:`~S3RequestType.DEFAULT`.
+                It is ignored for other types, since the operation is implicitly known.
+                See `S3 API documentation
+                <https://docs.aws.amazon.com/AmazonS3/latest/API/API_Operations_Amazon_Simple_Storage_Service.html>`_
+                for the canonical list of names.
+
                 This name is used to fill out details in metrics and error reports.
+                It also drives some operation-specific behavior.
+                If you pass the wrong name, you risk getting the wrong behavior.
+
+                For example, every operation except "GetObject" has its response checked
+                for error, even if the HTTP status-code was 200 OK
+                (see `knowledge center <https://repost.aws/knowledge-center/s3-resolve-200-internalerror>`_).
+                If you used the :attr:`~S3RequestType.DEFAULT` type to do
+                `GetObject <https://docs.aws.amazon.com/AmazonS3/latest/API/API_GetObject.html>`_,
+                but mis-named it "Download", and the object looked like XML with an error code,
+                then the request would fail. You risk logging the full response body,
+                and leaking sensitive data.
 
             recv_filepath (Optional[str]): Optional file path. If set, the
                 response body is written directly to a file and the
@@ -347,6 +380,20 @@ class S3Client(NativeResource):
 
             checksum_config (Optional[S3ChecksumConfig]): Optional checksum settings.
 
+            part_size (Optional[int]): Size, in bytes, of parts that files will be downloaded or uploaded in.
+                If not set, the part size configured for the client will be used.
+                Note: for :attr:`S3RequestType.PUT_OBJECT` request, client will adjust the part size to meet the service limits.
+                (max number of parts per upload is 10,000, minimum upload part size is 5 MiB)
+
+            multipart_upload_threshold (Optional[int]): The size threshold in bytes, for when to use multipart uploads.
+                This only affects :attr:`S3RequestType.PUT_OBJECT` request.
+                Uploads over this size will use the multipart upload strategy.
+                Uploads this size or less will use a single request.
+                If set, this should be at least `part_size`.
+                If not set, `part_size` adjusted by client will be used as the threshold.
+                If both `part_size` and `multipart_upload_threshold` are not set,
+                the values from `aws_s3_client_config` are used.
+
             on_headers: Optional callback invoked as the response received, and even the API request
                 has been split into multiple parts, this callback will only be invoked once as
                 it's just making one API request to S3.
@@ -401,6 +448,13 @@ class S3Client(NativeResource):
                         this is the final response's status code. If the operation
                         failed for another reason, None is returned.
 
+                    *   `did_validate_checksum` (bool):
+                            Was the server side checksum compared against a calculated checksum of the response body.
+                            This may be false even if :attr:`S3ChecksumConfig.validate_response` was set because
+                            the object was uploaded without a checksum, or downloaded differently from how it's uploaded.
+
+                    *   `checksum_validation_algorithm` (Optional[S3ChecksumAlgorithm]): The checksum algorithm used to validate the response.
+
                     *   `**kwargs` (dict): Forward-compatibility kwargs.
 
             on_progress: Optional callback invoked when part of the transfer is done to report the progress.
@@ -423,6 +477,8 @@ class S3Client(NativeResource):
             signing_config=signing_config,
             credential_provider=credential_provider,
             checksum_config=checksum_config,
+            part_size=part_size,
+            multipart_upload_threshold=multipart_upload_threshold,
             on_headers=on_headers,
             on_body=on_body,
             on_done=on_done,
@@ -458,6 +514,8 @@ class S3Request(NativeResource):
             signing_config=None,
             credential_provider=None,
             checksum_config=None,
+            part_size=None,
+            multipart_upload_threshold=None,
             on_headers=None,
             on_body=None,
             on_done=None,
@@ -468,14 +526,24 @@ class S3Request(NativeResource):
         assert callable(on_headers) or on_headers is None
         assert callable(on_body) or on_body is None
         assert callable(on_done) or on_done is None
+        assert isinstance(part_size, int) or part_size is None
+        assert isinstance(multipart_upload_threshold, int) or multipart_upload_threshold is None
+
+        if type == S3RequestType.DEFAULT and not operation_name:
+            raise ValueError("'operation_name' must be set when using S3RequestType.DEFAULT")
 
         super().__init__()
 
         self._finished_future = Future()
         self.shutdown_event = threading.Event()
 
-        checksum_algorithm = 0  # 0 means NONE in C
-        checksum_location = 0  # 0 means NONE in C
+        # C layer uses 0 to indicate defaults
+        if part_size is None:
+            part_size = 0
+        if multipart_upload_threshold is None:
+            multipart_upload_threshold = 0
+        checksum_algorithm = 0
+        checksum_location = 0
         validate_response_checksum = False
         if checksum_config is not None:
             if checksum_config.algorithm is not None:
@@ -509,6 +577,8 @@ class S3Request(NativeResource):
             checksum_algorithm,
             checksum_location,
             validate_response_checksum,
+            part_size,
+            multipart_upload_threshold,
             s3_request_core)
 
     @property
@@ -530,12 +600,11 @@ class S3ResponseError(awscrt.exceptions.AwsCrtError):
         headers (list[tuple[str, str]]): Headers from HTTP response.
         body (Optional[bytes]): Body of HTTP response (if any).
             This is usually XML. It may be None in the case of a HEAD response.
-        operation_name (Optional[str]): Name of the S3 operation that failed (if known).
+        operation_name: Name of the S3 operation that failed.
             For example, if a :attr:`~S3RequestType.PUT_OBJECT` fails
             this could be "PutObject", "CreateMultipartUpload", "UploadPart",
             "CompleteMultipartUpload", or others. For :attr:`~S3RequestType.DEFAULT`,
             this is the `operation_name` passed to :meth:`S3Client.make_request()`.
-            If the S3 operation name is unknown, this will be None.
         code (int): CRT error code.
         name (str): CRT error name.
         message (str): CRT error message.
@@ -623,7 +692,15 @@ class _S3RequestCore:
     def _on_shutdown(self):
         self._shutdown_event.set()
 
-    def _on_finish(self, error_code, status_code, error_headers, error_body, error_operation_name):
+    def _on_finish(
+            self,
+            error_code,
+            status_code,
+            error_headers,
+            error_body,
+            error_operation_name,
+            did_validate_checksum,
+            checksum_validation_algorithm):
         # If C layer gives status_code 0, that means "unknown"
         if status_code == 0:
             status_code = None
@@ -631,7 +708,6 @@ class _S3RequestCore:
         error = None
         if error_code:
             error = awscrt.exceptions.from_code(error_code)
-
             if isinstance(error, awscrt.exceptions.AwsCrtError):
                 if (error.name == "AWS_ERROR_CRT_CALLBACK_EXCEPTION"
                         and self._python_callback_exception is not None):
@@ -651,13 +727,21 @@ class _S3RequestCore:
             self._finished_future.set_exception(error)
         else:
             self._finished_future.set_result(None)
+
+        if checksum_validation_algorithm:
+            checksum_validation_algorithm = S3ChecksumAlgorithm(checksum_validation_algorithm)
+        else:
+            checksum_validation_algorithm = None
+
         if self._on_done_cb:
             self._on_done_cb(
                 error=error,
                 error_headers=error_headers,
                 error_body=error_body,
                 error_operation_name=error_operation_name,
-                status_code=status_code)
+                status_code=status_code,
+                did_validate_checksum=did_validate_checksum,
+                checksum_validation_algorithm=checksum_validation_algorithm)
 
     def _on_progress(self, progress):
         if self._on_progress_cb:
